%%% -*- indent-tabs-mode: nil -*-
%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000-2001
%%%
%%% Last change:
%%%   $Date: 2001/08/01 07:05:32 $ by $Author: afranke $
%%%   $Revision: 1.3 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% a simple HTTP server
%%

functor

import
   Error(exceptionToMessage messageToVirtualString printException)
   Open(socket:Socket)
   OS
   Pickle(load)
   Resolve
   System
   Type(ask)
   URL(make toString)

   File         at 'x-ozlib://mathweb/share/File.ozf'
   Internet     at 'x-ozlib://mathweb/share/Internet.ozf'
   XApplication at 'x-ozlib://mathweb/share/XApplication.ozf'
   XTime        at 'x-ozlib://mathweb/share/XTime.ozf'

   Buffered(socket:BufSocket) at 'x-ozlib://mathweb/html/Buffered.ozf'
   %% HTML('class':Html)         at 'x-ozlib://mathweb/html/Html.ozf'

   MIME_TYPES at 'x-ozlib://mathweb/httpd/mime.types.ozf.static'
   
   MimeTypes at 'x-ozlib://mathweb/httpd/MimeTypes.ozf'
   
   HttpCodes(codes: Codes)
   at 'x-ozlib://mathweb/httpd/HttpCodes.ozf'

   HttpRequest('class': HTTPRequest)
   at 'x-ozlib://mathweb/httpd/HttpRequest.ozf'

export
   readerSupport: HttpReaderSupport
   log:           Log
   
   connection: HttpConnection
   server:     HttpServer
   
prepare
   /*
   fun {DoConcat X Y} X#"\n"#Y end
   fun {Concat L} {FoldR L DoConcat nil} end
   */
   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
   fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end
   /*
   local
      fun {DoEnumerate Xs Sep}
         case Xs of nil then ''
         [] X|Xr then Sep#X#{DoEnumerate Xr Sep} end
      end
   in
      fun {Enumerate X|Xr Sep}
         X#{DoEnumerate Xr Sep}
      end
   end
   */

   fun {EscapeHtml S}
      case S of nil then nil
      [] C|Sr then
         case C
         of &< then &&|&l|&t|&;|{EscapeHtml Sr}
         [] &> then &&|&g|&t|&;|{EscapeHtml Sr}
         [] && then &&|&a|&m|&p|&;|{EscapeHtml Sr}
         [] &" then &&|&q|&u|&o|&t|&;|{EscapeHtml Sr}
         else C|{EscapeHtml Sr}
         end
      end
   end
   
define
   
   VS2S  = VirtualString.toString
   VS2BS = VirtualString.toByteString
   
   fun {ExceptionToVS Err}
      {Error.messageToVirtualString {Error.exceptionToMessage Err}}
   end
   
   AllMimeTypes = MIME_TYPES.mimeTypes
   
   ContentTypes = {Adjoin AllMimeTypes
                   mimeTypes('txt':  'text/plain'
                             'html': 'text/html'
                             'oza':  'application/x-oz-application')}
   
   fun {CondGetType Url Default}
      {MimeTypes.condGet ContentTypes Url Default}
   end   
   
   fun {Date}
      T = {XTime.gmTime} in
      T.wDay#", "#T.day#" "#T.mon#" "#T.year#" "#
      T.hour#":"#T.min#":"#T.sec#" "#"GMT"
   end
   
   ALLOW = "GET" %", HEAD, OPTIONS, TRACE"
   
   NOT_FOUND_GIF = "http://www.mathweb.org/~omega/www/ologo.gif"
   
   proc {GetUrl Resolver U ?VS}
      try
         case {Resolver.localize U}
         of old(F) then VS={File.loadBS F}
         [] new(F) then VS={File.loadBS F} {OS.unlink F}
         end
      catch _ then
         if {IsFree VS} then VS=false end
      end
   end
   
   local
      DefaultLevel = 1
   in
      class Log
         attr level: 1
         meth setLogLevel(Level)
            level <- Level
         end
         meth getLogLevel($)
            @level
         end
         meth isActiveLevel(Level $)
            @level>=Level
         end
         meth printInfo(VS level:Level<=DefaultLevel)
            if Log, isActiveLevel(Level $) then
               {System.printInfo VS}
            end
         end
         meth showInfo(VS level:Level<=DefaultLevel)
            Log, printInfo(VS#"\n" level:Level)
         end
         meth print(VS level:Level<=DefaultLevel)
            if Log, isActiveLevel(Level $) then
               {System.print VS}
            end
         end
         meth show(VS level:Level<=DefaultLevel)
            Log, print(VS level:Level)
            Log, printInfo("\n" level:Level)
         end
      end
   end

   %%
   %% generic HTTP/1.0 reader support
   %%
   class HttpReaderSupport
      %% getS reads a line, analogous to Open.text, getS
      meth GetS($ ?Newline)
         case {self read(list:$ size:1)} of nil then
            Log, showInfo('*** END OF TEXT ***' level:3)
            Newline=false nil
         [] [C] then
            if C==&\n then
               Log, showInfo('***\\n***' level:3)
               Newline=true nil
            else
               if {Char.isPrint C}
               then Log, printInfo([C] level:3)
               else Log, print({Char.toAtom C} level:3)
               end
               C|{self GetS($ ?Newline)}
            end
         end
      end
      meth getS($)
         S in
         if HttpReaderSupport, GetS(?S $)
         then S
         else false
         end
      end
      %% reads a line, raises an exception (instead of false, like getS)
      %% or returns the line without \r
      meth readLine($)
         case {self getS($)} of false then
            raise connectionClosed end
         [] S then
            case {Reverse S} of &\r|S1 then {Reverse S1}
            else S end
         end
      end
      meth readHeaders($)
         %% Header-Name: value
         %% returns a list of pairs (note: multiple headers are allowed)
         %% bug: headers over multiple lines are not recognized
         case HttpReaderSupport, readLine($) of nil then nil
         [] S then
            Feature Val in
            {String.token S &: ?Feature ?Val}
            ({StringToAtom {Map Feature Char.toLower}}#{Clear Val})
            | HttpReaderSupport, readHeaders($)
         end
      end
      meth readContent(Len ?Content)
         Tail RealLen in
         {self read(list:?Content tail:Tail size:Len len:?RealLen)}
         if RealLen == 0 orelse RealLen == Len
         then Tail=nil
         else {self readContent(Len-RealLen ?Tail)}
         end
      end
      meth readAll($)
         {self read(list:$ size:all)}
      end
   end
   
   %%
   %% reader for HTTP/1.0 request
   %%
   class HttpRequestReader from HttpReaderSupport
      meth InitialRequestLine(?Method ?RequestURI ?Protocol)
         %% GET /path/file.html HTTP/1.0
         S = HttpReaderSupport, readLine($)
         Words = {Filter {String.tokens S & }
                  fun {$ W} W\="" end}
      in
         case Words of [W1 W2 W3] then
            Method     = W1
            RequestURI = W2
            Protocol   = W3
         [] [W1 W2] then
            %% if not specified, assume HTTP/1.0
            Method     = W1
            RequestURI = W2
            Protocol   = "HTTP/1.0"
         else
            {self errBadRequest}
            raise close end
         end
      end
      meth readRequest(?Request)
         Request = {New HTTPRequest init}
         %% first line
         local
            ?Method ?RequestURI ?Protocol
         in
            {self InitialRequestLine(?Method ?RequestURI ?Protocol)}
            {Request setMethod(Method)}
            {Request setUri(RequestURI)}
            {Request setProtocol(Protocol)}
         end
         %% headers
         local
            ?Headers
         in 
            %% bug(?): multiple headers with the same name cause an error
            Headers = {List.toRecord headers {self readHeaders($)}}
            {Record.forAllInd Headers
             proc {$ HeaderNameA HeaderValVS}
                {Request addHeader(HeaderNameA HeaderValVS)}
             end}
         end
         %% (optional) content
         local
            ?Content
         in
            case {Request condGetHeader('content-length' false $)}
            of false then
               Content = nil
            [] ContentLengthVS then
               Len = {StringToInt {VS2S ContentLengthVS}} in
               Content={self readContent(Len $)}
            end
            {Request setContent(Content)}
         end 
      end                        
   end
   
   %%
   %% HTTP/1.0 responses
   %%
   class Self=HttpWriter from BufSocket

      attr
         Headers: nil
         contentType: 'unknown'
         streaming: false
         autoflush: false
         
      meth InitHeaders
         Headers <- nil
      end
      meth HeaderLine(VS)
         Headers <- @Headers#VS#'\r\n'
      end
      %%
      meth getContentType($)
         @contentType
      end
      meth openConsole
         skip
      end
      meth closeConsole
         skip
      end
      %%
      meth beginResponse(StatusCode <= 200
                         contentType: ContentType <= 'text/html'
                         protocol:    Protocol    <= "HTTP/1.0")
         /*
         %% in streaming mode, uncommenting this code will cause HTTP headers
         %% to show up in case of error messages
         if @streaming then
            streaming <- false
            autoflush <- false
         end
         */
         {self closeConsole}
         {self InitHeaders}
         case {CondSelect Codes StatusCode false} of false then
            raise unknownCode(StatusCode) end
         [] ReasonPhrase then
            {self HeaderLine(Protocol#" "#StatusCode#" "#ReasonPhrase)}
         end
         {self headerLine("Date"         {Date})}
         {self headerLine("Server"       "Ozhttpd/0.2 (Mozart)")}
         {self headerLine("Connection"   "close")}
         {self headerLine("Content-Type" ContentType)}
         case BufSocket, getAll($) of '' then skip
         [] VS then
            {System.showError "\n*** WARNING: Discarding buffer content:"}
            {System.showError {ByteString.make VS}}
         end
         %% hack to make plain text error messages readable in streaming html
         if @streaming andthen
            @contentType=='text/html' andthen
            ContentType=='text/plain' then
            {self write(vs:"<pre>\n")}
         end
         %% don't store the content type if headers have already been sent
         if @streaming==false then
            contentType <- ContentType
         end
      end
      meth writeHeaders
         {self HeaderLine("")}
         {self write(vs:@Headers)}
         {self InitHeaders}
      end
      meth beginContentStream(autoflush:Autoflush<=true)
         %% streaming content: use client flushBuffer
         {self writeHeaders}
         BufSocket, flushBuffer
         streaming <- true
         autoflush <- Autoflush
      end
      meth headerLine(HeaderName Value)
         if {IsVirtualString Value} then
            {self HeaderLine(HeaderName#': '#Value)}
         end
      end
      meth doctype
         {self write(vs:'<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">\n')}
      end
      meth baseURL($)
         self.server.baseURL
      end
      meth endResponse(close:Close<=false)
         if @streaming then
            streaming <- false
            autoflush <- false
         else
            BS = local
                    VS = BufSocket,getAll($) in
                    {ByteString.make VS}
                 end
         in
            {self headerLine("Content-Length" {ByteString.length BS})}
            {self writeHeaders}
            {self write(vs:BS)}
         end
         BufSocket, flushBuffer
         if Close then
            raise close end
         end
      end
      
      %%
      %% some standard responses in text/plain
      %%
      meth serverSignature(host: Host <= unit
                           port: Port <= unit)
         {self write(vs:"\n")}
         {self write(vs:"---\n")}
         {self write(vs:"MathWeb/0.2 Server")}
         if Host\=unit then
            {self write(vs: " at "#Host)}
            if Port\=unit then {self write(vs: " Port "#Port)} end
         end
         {self write(vs:"\n")}
      end
      meth errSignature(...) = M
         {self {Adjoin M serverSignature}}
      end
      meth errBadRequest(host: Host <= unit
                         port: Port <= unit)
         Msg = 
         "Your application sent a request that this server could not understand."
      in
         Self, beginResponse(400 contentType:'text/plain')
         {self write(vs:"\n")}
         {self write(vs:"Bad Request\n")}
         {self write(vs:"===========\n")}
         {self write(vs:"\n")}
         {self write(vs:Msg)}
         {self errSignature(host:Host port:Port)}
         Self, endResponse(close:true)
      end
      meth errNotFound(_)
         Msg  = "Sorry, but this page doesn't exist..." in
         Self, beginResponse(404 contentType:'text/plain')
         {self write(vs:Msg)}
         Self, endResponse(close:true)
      end
      meth errServerError(MsgVS
                          host: Host <= unit
                          port: Port <= unit)
         Self, beginResponse(500 contentType:'text/plain')
         {self write(vs:"\n")}
         {self write(vs:"Server Error\n")}
         {self write(vs:"============\n")}
         {self write(vs:"\n")}
         {self write(vs:"Sorry, an internal server error occurred "
                     #"while processing your request.")}
         {self write(vs:"\n")}
         if @contentType=='text/html'
         then {self write(vs:{EscapeHtml MsgVS})}
         else {self write(vs:MsgVS)}
         end
         {self write(vs:"\n")}
         {self serverSignature(host:Host port:Port)}
         Self, endResponse(close:true)
      end
      meth errMethodNotImplemented(Method RequestURI
                                   host:   Host       <= unit
                                   port:   Port       <= unit)
         Self, beginResponse(501 contentType:'text/plain')
         {self headerLine("Allow" ALLOW)}
         {self write(vs:"\n")}
         {self write(vs:"Method Not Implemented\n")}
         {self write(vs:"======================\n")}
         {self write(vs:"\n")}
         {self write(vs:Method#" to "#RequestURI#" not supported.\n")}
         {self write(vs:"\n")}
         {self serverSignature(host:Host port:Port)}
         Self, endResponse(close:true)
      end
      meth ok(VS contentType:ContentType<='text/html')
         Self, beginResponse(contentType:ContentType)
         %%{self headerLine("Accept-ranges" "bytes")}
         {self write(vs:VS)}
         Self, endResponse
      end
   end

   
   %%
   %% connection manager
   %%
   class HttpConnection
      from BufSocket /*Html*/ HttpRequestReader HttpWriter Log %Text
      feat
         remoteHost
         remotePort
         server
         
      meth doConnection(server:S remoteHost:H remotePort:P)
         Log, setLogLevel(S.logLevel)
         self.server     = S
         self.remoteHost = H
         self.remotePort = P
         try
            Log, showInfo("*** connection accepted from "#H#", port "#P
                          level:1)
            HttpConnection, ReadLoop
            {System.showInfo "after readloop"}
         catch E then
            if E\=connectionClosed andthen E\=close then
               {Error.printException E}
            end
         finally
            try BufSocket,shutDown
            catch Err then {Error.printException Err} end
            BufSocket,close
            Log, showInfo("*** connection closed ("#H#", port "#P#")"
                          level:1)
         end
      end
      meth ReadLoop
         Request = {self readRequest($)} in
         {self PrintRequest(Request)}
         try
            %%{System.showInfo " before do request "}
            {self DoRequest(Request)}
            %%{System.showInfo " do request finished"}
         catch E=close(...) then
            %%{System.showInfo " exception close in Readloop"}
            raise E end 
         [] E then
            {self errServerError({VS2S {ExceptionToVS E}})}
         end
         HttpConnection, ReadLoop
      end
      meth PrintRequest(Request)
         %% first line
         local
            Method     = {Request getMethodVS($)}
            RequestURI = {Request getUriVS($)}
            Protocol   = {Request getProtocolVS($)}
         in
            Log, showInfo("[read]: "#Method#" "#RequestURI#" "#Protocol
                          level:1)
         end
         %% headers
         local
            Headers = {Request getHeadersR($)}
         in
            {Record.forAllInd Headers
             proc {$ F V} Log, showInfo("[read]: "#F#":\t"#V level:2) end}
            Log, showInfo('' level:2)
         end
         %% content
         local
            ContentBS = {Request getContentBS($)}
         in
            if {ByteString.length ContentBS}\=0 then
               Log, showInfo("[read]: "#ContentBS level:2)
            end
         end
      end
      meth DoRequest(Request)
         {self.server handleRequest(Request self)}
         raise close end
      end
      
      %%
      %% GET method default implementation
      %%
      meth 'GET'(RequestURI...)
         U = {URL.make RequestURI} in
         {self getUrl(U resolver:self.server.resolver)}
      end
      meth getUrl(U resolver:Resolver<=Resolve)
         UrlVS = {URL.toString U} in
         Log, showInfo("Getting "#UrlVS level:1)
         case {GetUrl Resolver U} of false then
            {self errNotFound}
         [] VS then
            {self ok(VS type:{CondGetType UrlVS 'text/html'})}
         end
      end

      %%
      %% POST method default implementation
      %%
      meth 'POST'(RequestURI Content...)
         {self errMethodNotImplemented("POST" RequestURI)}
      end
      
      %%
      %% errors & responses
      %%
      meth errBadRequest
         HttpWriter, errBadRequest(host: self.server.host
                                   port: self.server.port)
      end
      meth errNotFound
         HttpWriter, errNotFound(self.server.'NOT_FOUND_GIF')
      end
      meth errServerError(MsgVS<="")
         HttpWriter, errServerError(MsgVS
                                    host: self.server.host
                                    port: self.server.port)
      end
      meth errMethodNotImplemented(Method RequestURI)
         HttpWriter, errMethodNotImplemented(Method RequestURI
                                             host: self.server.host
                                             port: self.server.port)
      end
      meth ok(ContentVS type:ContentType<='text/html')
         HttpWriter, ok(ContentVS contentType:ContentType)
      end
      meth applet(ContentVS)
         HttpWriter, ok(ContentVS contentType:'application/x-oz-application')
      end
      
      %%
      %% page output
      %%
      meth write(vs:VS)
         Log, printInfo("[write]: " level:2)
         Log, show({ByteString.make VS} level:2)
         BufSocket, write(vs:VS)
         if @autoflush then
            BufSocket, flushBuffer
         end
      end
   end

   %%
   %% handlers
   %%
   fun {StartsWith S Prefix}
      {List.isPrefix Prefix S}
   end
   fun {EndsWith S Suffix}
      {List.isPrefix {Reverse Suffix} {Reverse S}}
   end
   %%
   fun {PathToString P}
      {URL.toString url(path:P absolute:true)}
   end
   %%
   proc {DefaultHandler Request Connection}
      Method   = {Request getMethodS($)}
      Uri      = {Request getUriS($)}
      Protocol = {Request getProtocolS($)}
      Headers  = {Request getHeadersR($)}
   in
      case Method
      of "GET"  then
         {Connection 'GET'(Uri Protocol Headers)}
      [] "POST" then
         Content =  {Request getContentVS($)} in
         {Connection 'POST'(Uri Content Protocol Headers)}
         {Connection flushBuffer}
         {Delay 1000}
      else
         {Connection errMethodNotImplemented(Method Uri)}   
      end
   end
   proc {Skip0} skip end
   %%
   class HandlerRegistry

      prop
         locking

      feat
         counter
         
      attr
         Handlers: nil

      meth initHandlers
         self.counter = {New XApplication.counter init(0)}
      end
      
      meth handleRequest(Request Connection)
         RequestUri = {Request getUriR($)} in
         case HandlerRegistry, getHandler(RequestUri.path $)
         of false then
            {DefaultHandler Request Connection}
         [] Handler # PathInContext then
            UriInContext = {AdjoinAt RequestUri path PathInContext} in
            {Handler handle(UriInContext Request Connection)}
         end
      end

      meth getHandler(PathL $)
         %%{System.showInfo "getting handler for \""#{PathToString PathL}#"\""}
         lock
            fun {FindIn Hs}
               case Hs of nil then false
               [] (CP # Entry) | Hr then
                  %%{System.printInfo "matching \""#{PathToString CP}#"\" ..."}
                  if {List.isPrefix CP PathL}
                  then
                     %%{System.showInfo " succeeded"}
                     (Entry.1) # {List.drop PathL {Length CP}}
                  else
                     %%{System.showInfo " failed"}
                     {FindIn Hr}
                  end
               end
            end
         in
            {FindIn @Handlers}
         end
      end
      
      meth addHandler(ContextPathSpec Handler ?ID<=_
                      onRemove:OnRemove<=Skip0)
         %% make sure ContextPathSpec is a string
         CP  = {URL.toString {URL.make ContextPathSpec}}
         Len = {Length CP}
         %% check context path
         if CP==nil
            orelse {Member &, CP}
            orelse {StartsWith CP "/"}==false
            %% orelse {StartsWith CP "*"} % redundant
         then
            raise illegalContextSpec({ByteString.make CP}) end
         end
         %% strip trailing "/*" or "/"
         CP1 = if {EndsWith CP "/*"}    then {List.take CP Len-2}
               elseif Len=<1            then "/"
               elseif {EndsWith CP "/"} then {List.take CP Len-1}
               else CP
               end
         %% convert to normal form
         CP2 = {URL.make CP1}.path
         %% doublecheck result
         if {PathToString CP2}\=CP1 then
            raise illegalContextSpec({ByteString.make CP} pathtest) end
         end
         %% remove trailing "nil"
         CP3 = case {Reverse CP2} of nil|L then {Reverse L}
               else CP2 end
      in
         {Type.ask.'procedure/0' OnRemove}
         lock
            !ID = {self.counter next($)}
            Entry = entry(Handler id:ID onRemove:OnRemove)
         in
            Handlers <- (CP3 # Entry) | @Handlers
         end
      end
      
      meth removeHandler(ID)
         {Type.ask.int ID}
         lock
            fun {RemoveFrom Hs}
               case Hs of nil then
                  raise handlerNotFound(ID) end
               [] (Pair = (_ # Entry)) | Hr then
                  if Entry.id == ID then
                     OnRemove = {CondSelect Entry onRemove Skip0}
                     Done
                  in
                     thread {OnRemove} Done=unit end
                     {WaitOr Done {Alarm 10000}} % signal error ?
                     Hr
                  else
                     Pair | {RemoveFrom Hr}
                  end
               end
            end
            RemainingHandlers = {RemoveFrom @Handlers}
         in
            Handlers <- RemainingHandlers
         end
      end
      
      meth removeAllHandlers
         lock
            {ForAll @Handlers
             proc {$ _#Entry}
                OnRemove = {CondSelect Entry onRemove Skip0}
                Done
             in
                thread {OnRemove} Done=unit end
                {WaitOr Done {Alarm 10000}} % signal error?
             end}
         end
      end

      meth showHandlers
         lock
            {System.showInfo "Registered HTTP handlers:"}
            {ForAll @Handlers
             proc {$ ContextPath#Entry}
                {System.showInfo "ID="#Entry.id#", "#
                 "ContextPath="#{PathToString ContextPath}}
             end}
         end
      end
   end
   
   %%
   %% server
   %%  
   class HttpServer from Socket HandlerRegistry

      feat
         host
         port
         baseURL
         acceptClass
         resolver
         'NOT_FOUND_GIF'
         AcceptLoopThread
         Terminated
         logLevel
         
      meth init(port:P exit:?Exit<=_
                resolver:    Resolver    <= false
                notFoundGif: NotFoundGif <= NOT_FOUND_GIF
                acceptClass: AcceptClass <= HttpConnection
                logLevel:    LogLevel    <= 1)
         {Type.ask.int LogLevel}
         self.logLevel = LogLevel
         {Type.ask.'class' AcceptClass}
         self.acceptClass = AcceptClass
         self.resolver    = if Resolver == false then false
                            else {Resolve.make 'httpd' vs(Resolver#":=")} end
         Socket, init
         Socket, bind(takePort:P)
         Socket, listen
         self.host        = Internet.hostname
         self.port        = P
         self.baseURL     = {VS2BS "http://"#self.host#
                             if self.port\=80 then ":"#self.port else "" end
                             #"/"}
         self.'NOT_FOUND_GIF' = NotFoundGif
         HandlerRegistry, initHandlers
         HttpServer, StartAcceptLoop(?Exit)
      end

      meth StartAcceptLoop(?Exit)
         thread
            proc {AcceptLoop}
               H P M in
               %% Oz internal: Socket, accept expects a class!
               Socket, accept(acceptClass: self.acceptClass
                              accepted:    ?M
                              host:?H port:?P)
               thread  
                  {M doConnection(server:     self
                                  remoteHost: H
                                  remotePort: P)}
               end
               {AcceptLoop}
            end
         in
            try
               self.AcceptLoopThread = {Thread.this}
               {AcceptLoop}
            catch Err then
               case Err of system(kernel(terminate...)...) then skip
               else Exit=Err end
            finally
               if {IsFree Exit} then
                  try Exit=unit
                  catch _ then skip end
               end
               try
                  HttpServer, Close
                  self.Terminated = unit
               catch Err then
                  self.Terminated = Err
               end
            end
         end
      end
      
      meth Close
         try Socket, shutDown
         catch system(os(os "shutdown"...)...) then skip end
         Socket, close
      end
      
      meth close
         if {IsFree self.AcceptLoopThread} then
            try self.AcceptLoopThread=unit
            catch failure(...) then skip end
         end
         if self.AcceptLoopThread\=unit then
            try {Thread.terminate self.AcceptLoopThread}
            catch error(kernel(deadThread...)...) then skip end
            {Wait self.Terminated}
            if self.Terminated\=unit then
               raise self.Terminated end
            end
         end
         HandlerRegistry, removeAllHandlers
      end
   end
end


/*
class HtmlResponses
   from Html
                          
   meth serverSignature(host: Host <= unit
                        port: Port <= unit)
      Html, tag(hr)
      Html, startTag(address)
      {self write(vs: "MathWeb/0.2 Server")}
      if Host\=unit then
         {self write(vs: " at "#Host)}
         if Port\=unit then {self write(vs: " Port "#Port)} end
      end
      Html, endTag(address)
   end
   meth errBadRequest(host: Host <= unit
                      port: Port <= unit)
      Msg =
      "Your browser sent a request that this server could not understand."
   in
      {self beginResponse(400)}
      {self doctype}
      Html, tag(html(head(title("400 Bad Request"))
                     body(h1("Bad Request")
                          Msg p
                          %% "Explanatory Text" p
                          'self'(errSignature(host:Host port:Port))
                         )))
      {self endResponse}
   end
   meth errNotFound(GifUrl)
      Msg  = "Sorry, but this page doesn't exist..." in
      {self beginResponse(404)}
      {self doctype}
      Html, tag(html(head(title("Sorry."))
                     body(bgcolor:'#ffffff'
                          center(img(src:GifUrl)
                                 h1(Msg)))))
      {self endResponse}
   end
   meth errServerError(MsgVS
                       host: Host <= unit
                       port: Port <= unit)
      {self beginResponse(500)}
      {self doctype}
      Html, tag(html(head(title("500 Server Error"))
                     body(h1("Server Error")
                          "Sorry, an internal server error occurred "
                          #"while processing your request." p
                          pre({EscapeHtml MsgVS})
                          'self'(serverSignature(host:Host port:Port))
                         )))
      {self endResponse}
   end
   meth errMethodNotImplemented(Method RequestURI
                                host:   Host       <= unit
                                port:   Port       <= unit)
      {self beginResponse(501)}
      {self headerLine("Allow" ALLOW)}
      {self doctype}
      Html, tag(html(head(title("501 Method Not Implemented"))
                     body(h1("Method Not Implemented")
                          Method#" to "#RequestURI#" not supported." p
                          "Invalid method." p
                          'self'(serverSignature(host:Host port:Port))
                         )))
      {self endResponse}
   end
   meth ok(VS contentType:ContentType<='text/html')
      {self beginResponse(contentType:ContentType)}
      %%{self headerLine("Accept-ranges" "bytes")}
      {self write(vs:VS)}
      {self endResponse}
   end
   %%
   meth simpleResponse(Body title:Title<='')
      {self beginResponse(contentType:'text/html')}
      Html, tag(html(head(title(Title))#body(Body)))
      {self endResponse}
      {Delay 1000}
      raise close end
   end
end
*/
