%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date: 2001/07/20 10:28:35 $ by $Author: afranke $
%%%   $Revision: 1.2 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% a simple HTTP client
%%

functor
   
import
   Error(printException)
   Open(socket:Socket)
   System
   URL
   
   %% File      at 'x-ozlib://mathweb/share/File.ozf'
   %% Internet  at 'x-ozlib://mathweb/share/Internet.ozf'
   %% XTime     at 'x-ozlib://mathweb/share/XTime.ozf'

   Buffered(socket:BufSocket) at 'x-ozlib://mathweb/html/Buffered.ozf'
   %% HTML('class':Html)         at 'x-ozlib://mathweb/html/Html.ozf'
   
   Http      at 'x-ozlib://mathweb/httpd/Http.ozf'
   
export
   'class':     Client
   post:        POST
   
prepare
   /*
   fun {DoConcat X Y} X#"\n"#Y end
   fun {Concat L} {FoldR L DoConcat nil} end
   */
   fun {DropSpaces S} {List.dropWhile S Char.isSpace} end
   %% fun {Clear S} {Reverse {DropSpaces {Reverse {DropSpaces S}}}} end

   local
      fun {NotSpace C} {Char.isSpace C}==false end
   in
      proc {GetWord S ?W ?Sr}
         {List.takeDropWhile S NotSpace ?W ?Sr} 
      end
   end
   /*
   local
      fun {DoEnumerate Xs Sep}
         case Xs of nil then ''
         [] X|Xr then Sep#X#{DoEnumerate Xr Sep} end
      end
   in
      fun {Enumerate X|Xr Sep}
         X#{DoEnumerate Xr Sep}
      end
   end
   */
   /*
   Codes = {Record.map unit(200: "OK"
                            400: "Bad Request"
                            404: "Not Found"
                            301: "Moved Permanently" 
                            302: "Moved Temporarily" 
                            303: "See Other" % HTTP 1.1 only 
                            %% retrieve 'Location:' response header 
                            500: "Server Error"
                            501: "Method Not Implemented"
                           )
            ByteString.make}
   */
   
define
   /*
   fun {Date}
      T = {XTime.gmTime} in
      T.wDay#", "#T.day#" "#T.mon#" "#T.year#" "#
      T.hour#":"#T.min#":"#T.sec#" "#"GMT"
   end
   */

   VS2BS = ByteString.make

   HttpReaderSupport = Http.readerSupport
   Log               = Http.log
   
   %%
   %% reader for HTTP/1.0 responses
   %%
   
   class HttpResponseReader from HttpReaderSupport
      meth InitialResponseLine(?HTTPVersion ?StatusCode ?ReasonPhrase)
         %% HTTP/1.0 404 Not Found
         S Sr S2 in
         S            = {DropSpaces HttpReaderSupport,readLine($)}
         Sr           = {DropSpaces {GetWord S ?HTTPVersion $}}
         ReasonPhrase = {DropSpaces {GetWord Sr ?S2 $}}
         
         if {String.isInt S2} then
            StatusCode = {StringToInt S2}
         else
            {self errBadResponse(statusCode {VS2BS S2})}
         end
      end
      meth errBadResponse(...) = M
         raise error(M) end
      end
      meth readResponse(?HTTPVersion ?StatusCode ?ReasonPhrase
                        ?Headers ?Content)
         {self InitialResponseLine(?HTTPVersion ?StatusCode ?ReasonPhrase)}
         Headers = {List.toRecord headers {self readHeaders($)}}
         if {HasFeature Headers 'content-length'} then
            Len = {StringToInt Headers.'content-length'} in
            Content = {self readContent(Len $)}
         else
            Content = {self readAll($)}
         end
      end
   end
   
   %%
   %% writer for HTTP/1.0 requests
   %%
   class HttpRequestWriter from BufSocket
                       
      attr Headers: nil
         
      meth InitHeaders
         Headers <- nil
      end
      meth HeaderLine(VS)
         Headers <- @Headers#VS#'\r\n'
      end
      meth headerLine(HeaderName Value)
         {self HeaderLine(HeaderName#': '#Value)}
      end
      %%
      meth beginRequest(Method RequestURI headers:Headers <=unit())
         {self InitHeaders}
         {self HeaderLine(Method#" "#RequestURI#" "#"HTTP/1.0")}
         {Record.forAllInd Headers
          proc {$ F V} {self headerLine(F V)} end}
         case BufSocket, getAll($) of '' then skip
         [] VS then
            {System.showError "\n*** WARNING: Discarding buffer content:"}
            {System.showError {ByteString.make VS}}
         end
      end
      meth endRequest
         BS = local
                 VS = BufSocket,getAll($) in
                 {ByteString.make VS}
              end
      in
         if {ByteString.length BS}>0 then %% should be: Method == 'POST'
            {self headerLine("Content-Length" {ByteString.length BS})}
         end
         {self HeaderLine("")}
         {self write(vs:@Headers)}
         {self InitHeaders}
         {self write(vs:BS)}
         BufSocket, flushBuffer
      end
   end

   %%
   %% connection manager
   %%
   
   class Client
      from BufSocket /*Html*/ HttpResponseReader HttpRequestWriter Log %Text
      feat
         serverHost
         serverPort
         
      meth beginPOST(Url headers:Headers<=unit)
         U = {URL.make Url}
         HostS PortS
         case {String.tokens U.authority &:}
         of [H] then
            HostS = H
            PortS = "80"
         [] [H P] then
            HostS = H
            PortS = P
         end
         PathU = {Record.subtractList U [scheme authority]}
         Path  = {URL.toVirtualStringExtended PathU unit(info:true)}
      in
         self.serverHost = HostS
         self.serverPort = {StringToInt PortS}
         Socket, connect(host:self.serverHost port:self.serverPort)
         HttpRequestWriter, beginRequest('POST' Path headers:Headers)
      end
       
      meth endPOST
         HttpRequestWriter, endRequest
         Socket, shutDown(how:[send])
      end
      
      meth close
	 %%{System.printInfo "shutDown..."}
         try Socket, shutDown
         catch _ then skip end
         Socket, close
      end
      
      %% old:
      meth doConnection(server:S remoteHost:H remotePort:P)
         try
            Log, showInfo("*** connection accepted from "#H#", port "#P)
            Client, ReadLoop
         catch E then
            if E\=connectionClosed andthen E\=close then
               {Error.printException E}
            end
         finally
            try BufSocket,shutDown
            catch Err then {Error.printException Err} end
            BufSocket,close
            Log, showInfo("*** connection closed ("#H#", port "#P#")")
         end
      end
      meth ReadLoop
         Method RequestURI HTTPVersion Headers Content in
         {self readRequest(?Method ?RequestURI ?HTTPVersion ?Headers ?Content)}
         {self PrintRequest(Method RequestURI HTTPVersion Headers Content)}
         {self DoRequest(Method RequestURI HTTPVersion Headers Content)}
         Client, ReadLoop
      end   
      meth DoRequest(Method RequestURI HTTPVersion Headers Content)
         case Method
         of "GET"  then
            {self 'GET'(RequestURI HTTPVersion Headers)}
         [] "POST" then
            {self 'POST'(RequestURI Content HTTPVersion Headers)}
            BufSocket, flushBuffer
            {Delay 1000}
         else
            {self errMethodNotImplemented(Method RequestURI)}   
         end
         raise close end
      end
         
      meth PrintRequest(Method RequestURI HTTPVersion Headers Content)
         Log, showInfo("[read]: "#Method#" "#RequestURI#" "#HTTPVersion)
         {Record.forAllInd Headers
          proc {$ F V} Log, showInfo("[read]: "#F#":\t"#V) end}
         Log, showInfo('')
         if Content\=nil then
            Log, showInfo("[read]: "#Content)
         end
      end
      
      %% debug messages
      meth write(vs:VS)
         %% Log, printInfo("[write]: ")
         %% Log, show({ByteString.make VS})
         BufSocket, write(vs:VS)
      end
   end
   
   proc {POST Request ?Response}
      C = {New Client init}
      {C beginPOST(Request.url headers:{CondSelect Request headers unit})}
      {C write(vs:Request.content)}
      {C endPOST}
      HTTPVersion StatusCode ReasonPhrase Headers Content 
      {C readResponse(?HTTPVersion ?StatusCode ?ReasonPhrase
                      ?Headers ?Content)}
      {C close}
      {System.showInfo "closed."}
   in
      Response = response(httpVersion:  HTTPVersion
                          statusCode:   StatusCode
                          reasonPhrase: ReasonPhrase
                          headers:      Headers
                          content:      Content)
   end
end

