%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date: 2001/07/13 13:12:04 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% an interface for "mime.types" files
%%

functor

import
   Open(file text)
   Property(get)
   
export
   global:    GlobalTypes
   'local':   LocalTypes
   all:       AllTypes
   
   readFile:  ReadMimeTypes

   getSuffix: GetSuffix 
      
   get:       GetType
   condGet:   CondGetType
   
define

   %%
   %% parse "mime.types" files
   %%
   
   class TextFile from Open.text Open.file end
   
   local
      proc {SplitWord S ?W1 ?R}
         case S of nil then W1=nil R=nil
         [] C|Sr then
            if {Char.isSpace C}
            then W1=nil R=Sr 
            else W1=C|{SplitWord Sr $ ?R}
            end
         end
      end
      fun {SplitWords S}
         case {List.dropWhile S Char.isSpace} of nil then nil
         [] S1 then
            W1 R in
            {SplitWord S1 ?W1 ?R}
            W1|{SplitWords R}
         end
      end
      fun {ReadLines F}
         case {F getS($)} of false then nil
         [] Line then
            case {SplitWords Line}
            of nil then {ReadLines F}
            [] W1|Wr then (W1#Wr)|{ReadLines F}
            end
         end
      end
   in
      fun {ReadMimeTypes FileName}
         D = {NewDictionary}
         F = {New TextFile init(name:FileName)}
      in
         try
            {ForAll {ReadLines F}
             proc {$ MimeType#Extensions}
                T = {ByteString.make MimeType} in
                {ForAll Extensions
                 proc {$ Ext}
                    {Dictionary.put D {StringToAtom Ext} T}
                 end}
             end}
         finally
            {F close}
         end
         {Dictionary.toRecord mimeTypes D}
      end
   end

   %%
   %% mapping tables
   %%
   
   GlobalFile  = "/etc/mime.types"
   %%LocalFile   = {Property.get 'user.home'}#"/.mime.types"
   
   GlobalTypes = try {ReadMimeTypes GlobalFile}
                 catch _ then unit end
   
   LocalTypes  = unit
   %% hack; ~/.mime.types has a different syntax
   %% try {ReadMimeTypes LocalFile}
   %% catch _ then unit end
   
   AllTypes    = {Adjoin GlobalTypes LocalTypes}

   %%
   %% lookup
   %%

   VS2S = VirtualString.toString

   fun {GetSuffix Url}
      {Reverse {String.token {Reverse {VS2S Url}} &. $ _}}
   end
   fun {GetType Types Url}
      Types.{StringToAtom {GetSuffix Url}}
   end
   fun {CondGetType Types Url Default}
      {CondSelect Types {StringToAtom {GetSuffix Url}} Default}
   end
   
end


