%%% -*- indent-tabs-mode: nil -*-
%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date: 2001/07/20 15:40:35 $ by $Author: afranke $
%%%   $Revision: 1.2 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% an http server for mathweb
%%

functor

import
   OS(signal tmpnam unlink)
   Resolve(localize)
   System(showInfo showError)
   
   Internet(hostname)
   at 'x-ozlib://mathweb/share/Internet.ozf'

   XApplication(manager)
   at 'x-ozlib://mathweb/share/XApplication.ozf'
   
   Http(server)
   at 'x-ozlib://mathweb/httpd/Http.ozf'

   BrokerAccess(getBroker)
   at 'x-ozlib://mathweb/sys/BrokerAccess.ozf'
   
define
   AppName = 'httpd'
   App = {New XApplication.manager init}
   
   Spec = record(port(single char:&p type:int default:8080)
                 resolver(single char:&r type:string default:false)
                 default(single type:string default:false)
                )
   Args
   try
      Args = {App.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {System.showError AppName#": "#S}
      {App.exit 1}
   end

   fun {EscapeColon S}
      if {Member &: S} then
         Scheme in
         Scheme#"\\:"#{String.token S &: ?Scheme $}
      else S end
   end
   
   Resolver = if Args.resolver\=false then Args.resolver
              elseif Args.default\=false then
                 AddSlash = if {List.last Args.default}\=&/
                            then "/" else nil end
              in
                 "prefix=/="#{EscapeColon Args.default}#AddSlash
              else false
              end
   
   Server = {New Http.server init(port:     Args.port
                                  resolver: Resolver)}
   
   {System.showInfo Internet.hostname#":"#Args.port}

   ServiceID
   
   class HttpdService
      
      meth init(...)  skip end
      meth enter(...) skip end
      meth leave(...) skip end

      meth ping($) unit end
      
      meth getHost($)    Server.host    end
      meth getPort($)    Server.port    end
      meth getBaseURL($) Server.baseURL end
      
      meth addHandler(...)=Msg    {Server Msg} end
      meth removeHandler(...)=Msg {Server Msg} end

      meth showHandlers(...)=Msg  {Server Msg} end
      
      meth 'SHUTDOWN'
         {App.showInfo "*** httpd shutdown"}
         try
            %%{App.printInfo "*** unregistering from broker..."}
            {WaitOr {Alarm 10000}
             {Broker retire(ServiceID done:$)}}
            %%{App.showInfo " done"}
         catch Err then
            {App printExc(Err)}
         end
         try
            {App.printInfo "*** closing server..."}
            {Server close}
            {App.showInfo " done"}
         catch Err then
            {App printExc(Err)}
         end
      end
   end
   
   Httpd = {App newServer(HttpdService init $
                          detach: true
                          info:   "httpd")}

   Broker = {BrokerAccess.getBroker App Args}.send
   {Broker offer('HTTP_SERVER' Httpd ?ServiceID mode: pass_on)}
   
   {App.exit 0}
end
