%%% -*- indent-tabs-mode: nil -*-
%%%
%%% Author:
%%%   Andreas Franke <afranke@ags.uni-sb.de>
%%%
%%% Copyright:
%%%   Andreas Franke, 2000
%%%
%%% Last change:
%%%   $Date: 2001/07/13 13:12:05 $ by $Author: afranke $
%%%   $Revision: 1.1 $
%%%
%%% This file is part of MathWeb
%%%    http://www.mathweb.org/mathweb
%%%
%%% See the file "LICENSE" or
%%%    http://www.mathweb.org/mathweb/LICENSE.html
%%% for information on usage and redistribution
%%% of this file, and for a DISCLAIMER OF ALL
%%% WARRANTIES.
%%%

%%
%% a simple httpd proxy
%%

functor

import
   Application(getCmdArgs exit)
   OS(signal)
   System(showInfo showError)
   
   Internet(hostname)
   at 'x-ozlib://mathweb/share/Internet.ozf'
   
   Http(server: BaseServer)
   at 'x-ozlib://mathweb/httpd/Http.ozf'

   SampleHandlers
   at 'x-ozlib://mathweb/httpd/SampleHandlers.ozf'
   
define
   
   AppName = 'ozhttpd'
   
   Spec = record(port(single char:&p type:int default:8080)
                 resolver(single char:&r type:string default:false)
                 default(single type:string default:false)
                )
   Args
   try
      Args = {Application.getCmdArgs Spec}
   catch error(ap(usage S)...) then
      {System.showError AppName#": "#S}
      {Application.exit 1}
   end
   
   fun {EscapeColon S}
      if {Member &: S} then
         Scheme in
         Scheme#"\\:"#{String.token S &: ?Scheme $}
      else S end
   end
   
   Resolver = if Args.resolver\=false then Args.resolver
              elseif Args.default\=false then
                 AddSlash = if {List.last Args.default}\=&/
                            then "/" else nil end
              in
                 "prefix=/="#{EscapeColon Args.default}#AddSlash
              else false
              end
   
   Server = {New BaseServer init(port:     Args.port
                                 resolver: Resolver)}

   proc {AddHandler ContextPath Handler}
      ShowRemove = proc {$} {System.showInfo "remove "#ContextPath} end in
      {Server addHandler(ContextPath Handler onRemove:ShowRemove)}
   end
   {AddHandler "/http/*"   {New SampleHandlers.proxy init}}
   {AddHandler "/dump/*"   {New SampleHandlers.dump init}}
   {AddHandler "/applet/*" {New SampleHandlers.applet init}}
   
   {System.showInfo Internet.hostname#":"#Args.port}

   local
      T = {Thread.this}
      proc {Term SIGNAL}
         {System.showError '*** '#SIGNAL#' ***'}
         try {Thread.terminate T}
         catch _ then skip end
         {Delay 1000}
         {Application.exit 0}
      end
   in
      {OS.signal 'SIGHUP' Term}
      {OS.signal 'SIGINT' Term}
      {OS.signal 'SIGTERM' Term}
   end
   
   try
      {Thread.suspend {Thread.this}}
   finally
      {Server close}
      {Application.exit 0}
   end
end
